from amqpworker.easyqueue.message import AMQPMessage
from amqpworker.options import Actions


class RabbitMQMessage:
    def __init__(
            self,
            delivery_tag: int,
            amqp_message: AMQPMessage,
            on_success: Actions = Actions.ACK,
            on_exception: Actions = Actions.REQUEUE,
    ) -> None:
        self._delivery_tag = delivery_tag
        self._on_success_action = on_success
        self._on_exception_action = on_exception
        self._final_action = None
        self._amqp_message = amqp_message

    @property
    def body(self):
        return self._amqp_message.deserialized_data

    @property
    def serialized_data(self):
        return self._amqp_message.serialized_data

    def reject(self, requeue=True):
        """
        Marca essa mensagem para ser rejeitada. O parametro ``requeue`` indica se a mensagem será recolocada na fila original (``requeue=True``) ou será descartada (``requeue=False``).
        """
        self._final_action = Actions.REQUEUE if requeue else Actions.REJECT

    def requeue(self):
        self._final_action = Actions.REQUEUE_TAIL

    def accept(self):
        """
        Marca essa mensagem para ser confirmada (``ACK``) ao fim da execução do handler.
        """
        self._final_action = Actions.ACK

    def _process_action(self, action: Actions):
        if action == Actions.REJECT:
            self._amqp_message.reject(requeue=False)
        elif action == Actions.REQUEUE:
            self._amqp_message.reject(requeue=True)
        elif action == Actions.REQUEUE_TAIL:
            self._amqp_message.requeue()
        elif action == Actions.ACK:
            self._amqp_message.ack()

    def process_success(self):
        action = self._final_action or self._on_success_action
        return self._process_action(action)

    def process_exception(self):
        action = self._final_action or self._on_exception_action
        return self._process_action(action)



from typing import Generic, List, TypeVar

T = TypeVar("T")


class Bucket(Generic[T]):
    def __init__(self, size: int) -> None:
        self.size = size
        # fixme: Criar uma interface comum para as *Message
        # para substituir esse Any
        self._items: List[T] = []

    def is_full(self) -> bool:
        return len(self._items) == self.size

    def is_empty(self) -> bool:
        return len(self._items) == 0

    def put(self, item: T):
        if self.is_full():
            error_msg = f"Bucket is at full capacity: {self.size}"
            raise BucketFullException(error_msg)
        self._items.append(item)

    def pop_all(self) -> List[T]:
        _r = self._items
        self._items = []
        return _r

    @property
    def used(self) -> int:
        return len(self._items)


class BucketFullException(Exception):
    pass



import abc
import collections
from collections.abc import KeysView, ValuesView
from typing import (
    Any,
    Counter,
    Dict,
    ItemsView,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)

from pydantic import BaseModel, validator

from amqpworker.conf import settings
from amqpworker.easyqueue.queue import JsonQueue
from amqpworker.exceptions import InvalidConnection
from amqpworker.options import RouteTypes
from amqpworker.signals.base import Freezable


class Connection(BaseModel, abc.ABC):
    """
    Common ancestral for all Connection classes that auto generates a
    connection name and is responsible for keeping track of new connections on
    the ConnectionsMapping
    """

    route_type: RouteTypes
    name: Optional[str] = None


class ConnectionsMapping(Mapping[str, Connection], Freezable):
    """
    A mapping (Connection.name->Connection) of all available connections that
    also keeps a counter for each connection type
    """

    def __getitem__(self, k: str) -> Connection:
        return self._data[k]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __init__(self) -> None:
        Freezable.__init__(self)
        self._data: Dict[str, Connection] = {}
        self.counter: Counter[Type[Connection]] = collections.Counter()

    def __contains__(self, item):
        if isinstance(item, Connection):
            return item in self.values()
        return super(ConnectionsMapping, self).__contains__(item)

    def __setitem__(self, key: str, value: Connection) -> None:
        if self.frozen:
            raise RuntimeError(
                "You shouldn't change the state of ConnectionsMapping "
                "after App startup"
            )

        if key is None:
            key = id(value)

        if key in self:
            raise InvalidConnection(
                f"Invalid connection: `{value}`. "
                f"The name `{key}` already exists in {self.__class__.__name__}"
            )
        self._data[key] = value
        self.counter[value.__class__] += 1

    def __delitem__(self, key: str) -> None:
        if self.frozen:
            raise RuntimeError(
                "You shouldn't change the state of ConnectionsMapping "
                "after App startup"
            )
        del self._data[key]

    def add(self, connections: Iterable[Connection]) -> None:
        for conn in connections:
            self[conn.name] = conn  # type: ignore

    def with_type(self, route_type: RouteTypes) -> List["Connection"]:
        # todo: manter uma segunda estrutura de dados ou aceitar O(n) sempre que chamado?
        return [conn for conn in self.values() if conn.route_type == route_type]


_TYPE_COUNTER: Counter[Type[Connection]] = collections.Counter()

Message = Union[List, Dict]


class AMQPConnection(Connection):
    hostname: str
    port: int
    username: str
    password: str
    route_type = RouteTypes.AMQP_RABBITMQ
    prefetch: int = settings.AMQP_DEFAULT_PREFETCH_COUNT
    heartbeat: int = settings.AMQP_DEFAULT_HEARTBEAT
    name: Optional[str] = None
    connections: Dict[str, JsonQueue] = {}

    class Config:
        arbitrary_types_allowed = True

    @validator("connections", pre=True, always=True, check_fields=False)
    def set_connections(cls, v):
        return v or {}

    def __len__(self) -> int:
        return len(self.connections)

    def __iter__(self):
        return iter(self.connections)

    def __getitem__(self, key: str) -> JsonQueue:
        """
        Gets a JsonQueue instance for a given virtual host

        :param key: The virtual host of the connection
        :return: An instance of the connection
        """
        try:
            return self.connections[key]
        except KeyError:
            conn: JsonQueue = JsonQueue(
                host=self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
                virtual_host=key,
            )
            self.connections[key] = conn
            return conn

    def keys(self):
        return KeysView(self)

    def items(self):
        return ItemsView(self)

    def values(self):
        return ValuesView(self)

    def register(self, queue: JsonQueue) -> None:
        self.connections[queue.virtual_host] = queue

    def put(
            self,
            routing_key: str,
            data: Any = None,
            serialized_data: Union[str, bytes, None] = None,
            exchange: str = "",
            vhost: str = settings.AMQP_DEFAULT_VHOST,
            properties: Optional[dict] = None,
            mandatory: bool = False,
            immediate: bool = False,
    ):
        conn = self[vhost]
        return conn.put(
            routing_key=routing_key,
            data=data,
            serialized_data=serialized_data,
            exchange=exchange,
            properties=properties,
            mandatory=mandatory,
            immediate=immediate,
        )



import time
import traceback
from concurrent.futures import ThreadPoolExecutor
# from concurrent.futures.thread import _shutdown
from typing import Dict, Type, Union

from amqpstorm import AMQPError
from loguru import logger

from amqpworker import conf
from amqpworker.easyqueue.message import AMQPMessage
from amqpworker.easyqueue.queue import JsonQueue, QueueConsumerDelegate
from amqpworker.options import Events, Options
from amqpworker.routes import AMQPRoute
from .bucket import Bucket
from .rabbitmq import RabbitMQMessage
from .scheduled import ScheduledThreadPoolExecutor
from .utils import call


class Consumer(QueueConsumerDelegate):
    def __init__(
            self,
            route_info: Union[Dict, AMQPRoute],
            host: str,
            port: int,
            username: str,
            password: str,
            prefetch_count: int = 128,
            bucket_class: Type[Bucket] = Bucket[RabbitMQMessage],
    ) -> None:
        self.route = route_info
        self._handler = route_info["handler"]
        self._queue_name = route_info["routes"]
        self._route_options = route_info["options"]
        self.host = host
        self.vhost = route_info.get("vhost", "/")
        self.bucket = bucket_class(
            size=min(self._route_options["bulk_size"], prefetch_count)
        )
        self.running = False
        self.queue: JsonQueue = JsonQueue(
            host,
            username,
            password,
            port,
            virtual_host=self.vhost,
            delegate=self,
            prefetch_count=prefetch_count,
            logger=conf.logger,
            connection_fail_callback=self._route_options.get(
                Options.CONNECTION_FAIL_CALLBACK, None
            ),
        )

        self.pool = ThreadPoolExecutor(thread_name_prefix=f'submit-{"-".join(self._queue_name)}-',
                                       max_workers=self._route_options.get('max_workers', 8))
        self.multiple_queue_pool = ThreadPoolExecutor(thread_name_prefix='consumer-queue-',
                                                      max_workers=len(self._queue_name))
        self.clock = ScheduledThreadPoolExecutor(max_workers=1, name='flush-rabbit-')
        self.clock_task = None

    @property
    def queue_name(self) -> str:
        return self._queue_name

    def on_before_start_consumption(
            self, queue_name: str, queue: "JsonQueue"
    ):
        """
        Coroutine called before queue consumption starts. May be overwritten to
        implement further custom initialization.

        :param queue_name: Queue name that will be consumed
        :type queue_name: str
        :param queue: AsynQueue instanced
        :type queue: JsonQueue
        """
        pass

    def on_consumption_start(self, consumer_tag: str, queue: "JsonQueue"):
        """
        Coroutine called once consumption started.
        """
        pass

    def on_queue_message(self, msg: AMQPMessage):
        """
        Callback called every time that a new, valid and deserialized message
        is ready to be handled.
        """
        if not self.bucket.is_full():
            message = RabbitMQMessage(
                delivery_tag=msg.delivery_tag,
                amqp_message=msg,
                on_success=self._route_options[Events.ON_SUCCESS],
                on_exception=self._route_options[Events.ON_EXCEPTION],
            )
            self.bucket.put(message)

        if self.bucket.is_full():
            return self._flush_bucket_if_needed()

    def _flush_clocked(self, *args, **kwargs):
        try:
            self._flush_bucket_if_needed()
        except Exception as e:
            conf.logger.error(
                {
                    "type": "flush-bucket-failed",
                    "dest": self.host,
                    "retry": True,
                    "exc_traceback": traceback.format_exc(),
                }
            )

    def _flush_bucket_if_needed(self):

        if not self.bucket.is_empty():
            all_messages = self.bucket.pop_all()
            try:
                conf.logger.debug(
                    {
                        "event": "bucket-flush",
                        "bucket-size": len(all_messages),
                        "handler": self._handler.__name__,
                    }
                )
                if self.running:
                    rv = self._handler(all_messages)
                    for _ in self.pool.map(call, [m.process_success for m in all_messages]):
                        pass
                    return rv
                else:
                    for m in all_messages:
                        m.process_exception()
            except AMQPError:
                raise
            except Exception as e:
                traceback.print_exc()
                if self.running:
                    for _ in self.pool.map(call, [m.process_exception for m in all_messages]):
                        pass
                else:
                    for m in all_messages:
                        m.process_exception()
                raise e

    # def on_queue_error(self, body, delivery_tag, error, queue):
    #     """
    #     Callback called every time that an error occurred during the validation
    #     or deserialization stage.
    #
    #     :param body: unparsed, raw message content
    #     :type body: Any
    #     :param delivery_tag: delivery_tag of the consumed message
    #     :type delivery_tag: int
    #     :param error: THe error that caused the callback to be called
    #     :type error: MessageError
    #     :type queue: JsonQueue
    #     """
    #     conf.logger.error(
    #         {
    #             "parse-error": True,
    #             "exception": "Error: not a JSON",
    #             "original_msg": body,
    #         }
    #     )
    #     try:
    #         queue.ack(delivery_tag=delivery_tag)
    #     except AMQPError as e:
    #         self._log_exception(e)

    def on_message_handle_error(self, handler_error: Exception, **kwargs):
        """
        Callback called when an uncaught exception was raised during message
        handling stage.

        :param handler_error: The exception that triggered
        :param kwargs: arguments used to call the coroutine that handled
        the message
        """
        self._log_exception(handler_error)

    def on_connection_error(self, exception: Exception):
        """
        Called when the connection fails
        """
        self._log_exception(exception)

    def _log_exception(self, exception):
        current_exception = {
            "exc_message": str(exception),
            "exc_traceback": traceback.format_exc(),
        }
        conf.logger.error(current_exception)

    def consume_all_queues(self, queue: JsonQueue):

        # for r in self.multiple_queue_pool.map(queue.consume, self._queue_name, repeat(self)):
        #     pass
        for queue_name in self._queue_name:
            # Por enquanto não estamos guardando a consumer_tag retornada
            # se precisar, podemos passar a guardar.
            conf.logger.debug(
                {"queue": queue_name, "event": "start-consume"}
            )
            queue.consume(queue_name=queue_name, pool=self.multiple_queue_pool, delegate=self)

    def keep_runnig(self):
        return self.running

    def stop(self):
        logger.debug('stop consumer')
        self.running = False
        self.queue.connection.close()
        self.pool.shutdown(False)
        self.multiple_queue_pool.shutdown(False)
        self.clock.stop(None)

    def start(self):
        self.running = True
        if not self.clock_task:
            seconds = self._route_options.get(
                Options.BULK_FLUSH_INTERVAL, conf.settings.FLUSH_TIMEOUT
            )
            self.clock.schedule_at_fixed_rate(self._flush_clocked, seconds, seconds)
            self.clock.start(None)
            self.clock_task = self._flush_clocked
        while self.keep_runnig():
            if not self.queue.connection.has_channel_ready():
                try:
                    self.consume_all_queues(self.queue)
                except RuntimeError as e:
                    traceback.print_exc()
                    if self.multiple_queue_pool._shutdown:  # or _shutdown:
                        conf.logger.info('app shutdown')
                    # if 'interpreter shutdown' in str(e):
                    #     return
                    else:
                        raise
                except KeyboardInterrupt:
                    self.stop()
                except Exception as e:

                    conf.logger.error(
                        {
                            "type": "connection-failed",
                            "dest": self.host,
                            "retry": True,
                            "exc_traceback": traceback.format_exc(),
                        }
                    )
            time.sleep(1)



from enum import Enum, auto
from typing import List


class AutoNameEnum(str, Enum):
    def _generate_next_value_(  # type: ignore
            name: str, start: int, count: int, last_values: List[str]
    ) -> str:
        return name.lower()


class Options(AutoNameEnum):
    BULK_SIZE = auto()
    BULK_FLUSH_INTERVAL = auto()
    MAX_CONCURRENCY = auto()
    CONNECTION_FAIL_CALLBACK = auto()


class Actions(AutoNameEnum):
    ACK = auto()
    REJECT = auto()
    REQUEUE = auto()
    REQUEUE_TAIL = auto()


class Events(AutoNameEnum):
    ON_SUCCESS = auto()
    ON_EXCEPTION = auto()


class DefaultValues:
    MAX_SUBMIT_WORKER_SIZE = 8
    BULK_SIZE = 1
    BULK_FLUSH_INTERVAL = 60
    ON_SUCCESS = Actions.ACK
    ON_EXCEPTION = Actions.REQUEUE_TAIL
    RUN_EVERY_MAX_CONCURRENCY = 1


class RouteTypes(AutoNameEnum):
    AMQP_RABBITMQ = auto()
    HTTP = auto()



import abc
from collections import UserDict
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Type,
    Union,
)

from cached_property import cached_property
from pydantic import BaseModel, Extra, root_validator, validator

from amqpworker import conf
from amqpworker.connections import AMQPConnection, Connection
from amqpworker.options import Actions, DefaultValues, RouteTypes

RouteHandler = Callable[..., Coroutine]


class Model(BaseModel, abc.ABC):
    """
    An abstract pydantic BaseModel that also behaves like a Mapping
    """

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError as e:
            raise KeyError from e

    def __setitem__(self, key, value):
        try:
            return self.__setattr__(key, value)
        except (AttributeError, ValueError) as e:
            raise KeyError from e

    def __eq__(self, other):
        if isinstance(other, dict):
            return self.dict() == other
        return super(Model, self).__eq__(other)

    def __len__(self):
        return len(self.__fields__)

    def keys(self):
        return self.__fields__.keys()

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default


class _RouteOptions(Model):
    pass


class Route(Model, abc.ABC):
    """
    An abstract Model that acts like a route factory
    """

    type: RouteTypes
    handler: Any
    routes: List[str]
    connection: Optional[Connection]
    options: _RouteOptions = _RouteOptions()

    @staticmethod
    def factory(data: Dict) -> "Route":
        try:
            type_ = data.pop("type")
        except KeyError as e:
            raise ValueError("Routes must have a type") from e

        if type_ == RouteTypes.HTTP:
            raise ValueError(f"'{type_}' is an invalid RouteType.")
        if type_ == RouteTypes.AMQP_RABBITMQ:
            return AMQPRoute(**data)
        raise ValueError(f"'{type_}' is an invalid RouteType.")


class AMQPRouteOptions(_RouteOptions):
    bulk_size: int = DefaultValues.BULK_SIZE
    max_workers: int = DefaultValues.MAX_SUBMIT_WORKER_SIZE
    bulk_flush_interval: int = DefaultValues.BULK_FLUSH_INTERVAL
    on_success: Actions = DefaultValues.ON_SUCCESS
    on_exception: Actions = DefaultValues.ON_EXCEPTION
    connection_fail_callback: Optional[
        Callable[[Exception, int], Coroutine]
    ] = None
    connection: Optional[Union[AMQPConnection, str]]

    class Config:
        arbitrary_types_allowed = False
        extra = Extra.forbid


class AMQPRoute(Route):
    type: RouteTypes = RouteTypes.AMQP_RABBITMQ
    vhost: str = conf.settings.AMQP_DEFAULT_VHOST
    connection: Optional[AMQPConnection]
    options: AMQPRouteOptions


class RoutesRegistry(UserDict):
    def _get_routes_for_type(self, route_type: Type) -> Iterable:
        return tuple((r for r in self.values() if isinstance(r, route_type)))

    @cached_property
    def amqp_routes(self) -> Iterable[AMQPRoute]:
        return self._get_routes_for_type(AMQPRoute)

    def __setitem__(self, key: RouteHandler, value: Union[Dict, Route]):
        if not isinstance(value, Route):
            route = Route.factory({"handler": key, **value})
        else:
            route = value

        super(RoutesRegistry, self).__setitem__(key, route)

    def add_route(self, route: Route) -> None:
        self[route.handler] = route

    def add_amqp_route(self, route: AMQPRoute) -> None:
        self[route.handler] = route

    def route_for(self, handler: RouteHandler) -> Route:
        return self[handler]



import logging
from typing import List, Optional

from loguru import logger as _logger
from pydantic import BaseSettings

from amqpworker.options import DefaultValues

INFINITY = float("inf")


class Settings(BaseSettings):
    LOGLEVEL: str = "ERROR"

    AMQP_DEFAULT_VHOST: str = "/"
    AMQP_DEFAULT_PREFETCH_COUNT: int = 128
    AMQP_DEFAULT_HEARTBEAT: int = 60

    HTTP_HOST: str = "127.0.0.1"
    HTTP_PORT: int = 8080

    FLUSH_TIMEOUT: int = DefaultValues.BULK_FLUSH_INTERVAL

    # metrics
    METRICS_NAMESPACE: str = "amqpworker"
    METRICS_APPPREFIX: Optional[str]
    METRICS_ROUTE_PATH: str = "/metrics"
    METRICS_ROUTE_ENABLED: bool = True
    METRICS_DEFAULT_HISTOGRAM_BUCKETS_IN_MS: List[float] = [
        10,
        50,
        100,
        200,
        500,
        1000,
        5000,
        INFINITY,
    ]

    class Config:
        allow_mutation = False
        env_prefix = "AMQPWORKER_"


settings = Settings()

loglevel = getattr(logging, settings.LOGLEVEL, logging.INFO)
_logger.level( settings.LOGLEVEL)
logger = _logger



import abc
import threading

from collections import UserList


class Freezable(metaclass=abc.ABCMeta):
    def __init__(self):
        self._frozen = False

    @property
    def frozen(self) -> bool:
        return self._frozen

    def freeze(self):
        self._frozen = True


class Signal(UserList, threading.Event):
    def __init__(self, owner: Freezable):
        UserList.__init__(self)
        threading.Event.__init__(self)
        self._owner = owner
        self.frozen = False

    def __repr__(self):
        return "<Signal owner={}, frozen={}, {!r}>".format(
            self._owner, self.frozen, list(self)
        )

    def send(self, *args, **kwargs):
        """
        Sends data to all registered receivers.
        """
        if self.frozen:
            raise RuntimeError("Cannot send on frozen signal.")

        for receiver in self:
            receiver(*args, **kwargs)

        self.frozen = True
        self._owner.freeze()
        self.set()



import threading
from typing import Callable, Optional, Union
from amqpstorm.base import Stateful
from amqpstorm import Connection, Channel, AMQPError

OnErrorCallback = Union[
    None, Callable[[Exception], None]
]


class AMQPConnection:
    def __init__(
            self,
            host: str,
            username: str,
            password: str,
            port: int = 5672,
            heartbeat: int = 60,
            virtual_host: str = "/",
            on_error: OnErrorCallback = None,
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.virtual_host = virtual_host
        self.heartbeat = heartbeat
        self._on_error = on_error
        self._connection_lock = threading.Lock()
        self.channel: Optional[Channel] = None
        self._protocol: Optional[Connection] = None

    @property
    def connection_parameters(self):
        return {
            "hostname": self.host,

            "port": self.port,
            "username": self.username,
            "password": self.password,
            "virtual_host": self.virtual_host,
            # "on_error": self._on_error,
            "heartbeat": self.heartbeat,
        }

    @property
    def is_connected(self) -> bool:
        return self._protocol and self._protocol.current_state == Stateful.OPEN

    def has_channel_ready(self):
        return self.channel and self.channel.is_open

    def close(self) -> None:
        if not self.is_connected:
            return None
        self._protocol.close()

    def _connect(self) -> None:
        with self._connection_lock:
            if self.is_connected and self.has_channel_ready():
                return

            try:
                if self._protocol:
                    self.channel = self._protocol.channel()
                    return
            except AMQPError as e:
                # Se não conseguirmos pegar um channel novo
                # a conexão atual deve mesmo ser renovada e isso
                # será feito logo abaixo.
                pass

            self._protocol = Connection(**self.connection_parameters)

            self.channel = self._protocol.channel()



class InvalidRoute(ValueError):
    """
    Defines an invalid route definition condition.
    """


class InvalidConnection(ValueError):
    """
    Defines an invalid connection definition condition.
    """



import logging
from asyncio import Task
from concurrent.futures import Future
from typing import TYPE_CHECKING, List

from loguru import logger

from amqpworker.connections import AMQPConnection
from amqpworker.consumer import Consumer
from amqpworker.options import RouteTypes
from amqpworker.signals.handlers.base import SignalHandler

if TYPE_CHECKING:  # pragma: no cover
    from amqpworker.app import App  # noqa: F401


class RabbitMQ(SignalHandler):
    def shutdown(self, app: "App"):
        logger.debug('shutdown rabbit consumers')
        if RouteTypes.AMQP_RABBITMQ in app:
            for consumer in app[RouteTypes.AMQP_RABBITMQ]["consumers"]:
                logger.debug(f'stop {consumer.host}')
                consumer.stop()
                logger.debug(f'stopped {consumer.host}')

    def startup(self, app: "App") -> List[Future]:
        tasks = []

        app[RouteTypes.AMQP_RABBITMQ]["consumers"] = []
        for route_info in app.routes_registry.amqp_routes:
            conn: AMQPConnection = app.get_connection_for_route(route_info)

            consumer = Consumer(
                route_info=route_info,
                host=conn.hostname,
                port=conn.port,
                username=conn.username,
                password=conn.password,
                prefetch_count=conn.prefetch,
            )
            app[RouteTypes.AMQP_RABBITMQ]["consumers"].append(consumer)
            conn.register(consumer.queue)
            task = app.loop.submit(consumer.start)

            tasks.append(task)

        return tasks
